#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2015 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import logging
    import logging.handlers
    from   nlmanager.nllistener import NetlinkManagerWithListener
    import nlmanager.nlpacket
    import os
    import Queue
    import select
    import socket
    import struct
    import threading
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class ClagNl():

    # Event code constants
    _GET_ALL_ADDRESSES      = 1
    _GET_ALL_LINKS          = 2
    _GET_ALL_NEIGHBORS      = 3
    _GET_ALL_MACS           = 4
    _GET_ALL_ROUTES         = 5
    _SHUTDOWN               = 6
    _SERVICE_NETLINK_QUEUE  = 'SERVICE_NETLINK_QUEUE'

    def __init__(self, Daemon, Log, Intf):
        self.Daemon = Daemon
        self.Log = Log
        self.Intf = Intf
        self.FdbSync = None

        # Data shared by all the threads
        self.keep_going = True

        self.nlm = None
        self.netlinkThread = threading.Thread(None, self.NetlinkThreadT, "NetlinkThreadT")
        self.netlinkThread.daemon = True
        self.netlinkThread.start()

    def __del__(self):
        # Signal the update thread to wake up and exit
        self.SignalThreads()
        self.JoinThreads()

    def SignalThreads(self):
        self.keep_going = False
        if self.nlm:
            self.nlm.workq.put((ClagNl._SHUTDOWN, None))
            self.nlm.listener.shutdown_event.set()

    def JoinThreads(self):
        if self.nlm:
            self.nlm.listener.join()
        self.netlinkThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_NETLINK, logStr)

    def NetlinkThreadT(self):
        try:
            self.NetlinkThread()
        except:
            self.Daemon.DumpTraceback()

    def SetFdbSync(self, FdbSync):
        self.FdbSync = FdbSync
        if self.nlm is not None:
            self.nlm.FdbSync = FdbSync

    def DumpMacs(self):
        if self.nlm is not None:
            self.nlm.workq.put((ClagNl._GET_ALL_MACS, None))

    def mac_int_to_str(self, mac_int):
        """
        Return an integer in MAC string format: xx:xx:xx:xx:xx:xx
        """
        return ':'.join(("%012x" % mac_int)[i:i+2] for i in range(0, 12, 2))

    def NetlinkThread(self):
        """
        This function operates as an independent thread of the script.
        """

        self.log_debug("Beginning execution of the thread NetlinkThread")

        # These are the groups we are interesting in listening to.
        groups = nlmanager.nlpacket.RTMGRP_LINK | nlmanager.nlpacket.RTMGRP_NEIGH

        # Override the nlmanager's mac_int_to_str function to print the MACs
        # like xx:xx:xx:xx:xx:xx instead of xxxx.xxxx.xxxx
        nlmanager.nlpacket.mac_int_to_str = self.mac_int_to_str

        # Instantiate a netlink manager with a listener.
        self.nlm = ClagNlmWListener(groups, self.Log, self.Intf)
        self.nlm.FdbSync = self.FdbSync

        # Enable any message debugging here.
        #self.nlm.debug_neighbor(True)
        #self.nlm.debug_link(True)

        # Ignore non-bridge address family message for neighbors
        self.nlm.filter_by_address_family(True, 'whitelist', nlmanager.nlpacket.RTM_NEWNEIGH, socket.AF_BRIDGE)
        self.nlm.filter_by_address_family(True, 'whitelist', nlmanager.nlpacket.RTM_DELNEIGH, socket.AF_BRIDGE)

        # Prime the pump by getting all interfaces and MACs
        self.nlm.workq.put((ClagNl._GET_ALL_LINKS, None))

        # This is the main thread loop. 
        while self.keep_going:

            try:
                (event, options) = self.nlm.workq.get(True, 60*60*24) # If timeout is None, signals are blocked
            except Queue.Empty:
                continue

            if event == ClagNl._GET_ALL_ADDRESSES:
                self.nlm.get_all_addresses()
            elif event == ClagNl._GET_ALL_LINKS:
                self.nlm.get_all_links()
                self.nlm.get_all_br_links()
            elif event == ClagNl._GET_ALL_NEIGHBORS:
                self.nlm.get_all_neighbors()
            elif event == ClagNl._GET_ALL_MACS:
                self.nlm.get_all_macs()
            elif event == ClagNl._GET_ALL_ROUTES:
                self.nlm.get_all_routes()
            elif event == ClagNl._SERVICE_NETLINK_QUEUE:
                self.nlm.service_netlinkq()
            elif event == ClagNl._SHUTDOWN:
                self.log_debug("NetlinkThread: shutting things down")
            else:
                raise Exception("Unsupported workq event %s" % event)

        self.log_debug("Finished execution of the thread NetlinkThread")


class ClagNlmWListener(NetlinkManagerWithListener):
    """
    The main job of this class is to override the functions of the
    NetlinkManagerWithListener class to do the things clag needs when items are
    placed on the netlink queue.
    """
    def __init__(self, groups, Log, Intf):
        use_color = Log.getHandlerStr() != "syslog"
        super(ClagNlmWListener, self).__init__(groups, use_color=use_color)
        self.Log = Log
        self.log = Log.logger
        self.Intf = Intf
        self.FdbSync = None

    def rx_rtm_newlink(self, msg):
        self.Intf.AddInterface(self, msg)
        #self.log.debug("RXed RTM_NEWLINK seq %d, pid %d, %d bytes, for %s, state %s" %
        #               (msg.seq, msg.pid, msg.length, msg.get_attribute_value(msg.IFLA_IFNAME), "up" if msg.is_up() else "down"))

    def rx_rtm_dellink(self, msg):
        self.Intf.DelInterface(self, msg.ifindex, msg.family)
        #self.log.debug("RXed RTM_DELLINK seq %d, pid %d, %d bytes, for %s, state %s" %
        #               (msg.seq, msg.pid, msg.length, msg.get_attribute_value(msg.IFLA_IFNAME), "up" if msg.is_up() else "down"))

    def rx_rtm_newaddr(self, msg):
        self.log.debug("RX RTM_NEWADDR for %s/%d on %d" %
                       (msg.get_attribute_value(msg.IFA_ADDRESS), msg.prefixlen, msg.ifindex))

    def rx_rtm_deladdr(self, msg):
        self.log.debug("RX RTM_DELADDR for %s/%d on %d" %
                       (msg.get_attribute_value(msg.IFA_ADDRESS), msg.prefixlen, msg.ifindex))

    def rx_rtm_newneigh(self, msg):
        """
        For new dynamic MAC addresses, enqueue them for processing by the FDB sync code.
        """
        if msg.family == socket.AF_BRIDGE and msg.state & msg.NUD_REACHABLE:
            k_mac  = msg.get_attribute_value(msg.NDA_LLADDR)
            k_vlan = msg.get_attribute_value(msg.NDA_VLAN)
            k_bif  = self.Intf.GetIfName(msg.ifindex)
            k_vni  = msg.get_attribute_value(msg.NDA_VNI)
            k_dst  = msg.get_attribute_value(msg.NDA_DST)
            k_mstr = self.Intf.GetIfName(msg.get_attribute_value(msg.NDA_MASTER))
            if self.FdbSync is not None:
                self.FdbSync.fdbWorkQueue.put((self.FdbSync._WORK_QUEUE_ADD, (k_mac, k_vlan, k_bif, k_vni, k_dst, k_mstr)))
                self.log.debug("Adding MAC address %s vlan %s bif %s vni %s dst %s master %s" % 
                               (k_mac, str(k_vlan), str(k_bif), str(k_vni), str(k_dst), str(k_mstr)))

    def rx_rtm_delneigh(self, msg):
        """
        For deleted dynamic MAC addresses, enqueue them for processing by the FDB sync code.
        """
        if msg.family == socket.AF_BRIDGE and msg.state & (msg.NUD_STALE | msg.NUD_REACHABLE):
            k_mac  = msg.get_attribute_value(msg.NDA_LLADDR)
            k_vlan = msg.get_attribute_value(msg.NDA_VLAN)
            k_bif  = self.Intf.GetIfName(msg.ifindex)
            k_vni  = msg.get_attribute_value(msg.NDA_VNI)
            k_dst  = msg.get_attribute_value(msg.NDA_DST)
            k_mstr = self.Intf.GetIfName(msg.get_attribute_value(msg.NDA_MASTER))
            if self.FdbSync is not None:
                self.FdbSync.fdbWorkQueue.put((self.FdbSync._WORK_QUEUE_DELETE, (k_mac, k_vlan, k_bif, k_vni, k_dst, k_mstr)))
                self.log.debug("Deleting MAC address %s vlan %s bif %s vni %s dst %s master %s" % 
                               (k_mac, str(k_vlan), str(k_bif), str(k_vni), str(k_dst), str(k_mstr)))

    def service_netlinkq(self):
        msg_count = {}
        processed = 0

        for (msgtype, length, flags, seq, pid, data) in self.netlinkq:
            processed += 1

            # If this is a reply to a TX message that debugs were enabled for then debug the reply
            if (seq, pid) in self.debug_seq_pid:
                debug = True
            else:
                debug = self.debug_this_packet(msgtype)

            if msgtype == nlmanager.nlpacket.RTM_NEWLINK or msgtype == nlmanager.nlpacket.RTM_DELLINK:
                msg = nlmanager.nlpacket.Link(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            elif msgtype == nlmanager.nlpacket.RTM_NEWADDR or msgtype == nlmanager.nlpacket.RTM_DELADDR:
                msg = nlmanager.nlpacket.Address(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            elif msgtype == nlmanager.nlpacket.RTM_NEWNEIGH or msgtype == nlmanager.nlpacket.RTM_DELNEIGH:
                msg = nlmanager.nlpacket.Neighbor(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            elif msgtype == nlmanager.nlpacket.RTM_NEWROUTE or msgtype == nlmanager.nlpacket.RTM_DELROUTE:
                msg = nlmanager.nlpacket.Route(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            else:
                self.log.warning('RXed unknown netlink message type %s' % msgtype)
                continue

            msg.decode_packet(length, flags, seq, pid, data)

            if not self.filter_permit(msg):
                continue

            if debug:
                msg.dump()

            # Only used for printing debugs about how many we RXed of each type
            if msg.msgtype not in msg_count:
                msg_count[msg.msgtype] = 0
            msg_count[msg.msgtype] += 1

            # Call the appropriate handler method based on the msgtype.  The handler
            # functions are defined in our child class.
            if msg.msgtype == nlmanager.nlpacket.RTM_NEWLINK:
                self.rx_rtm_newlink(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELLINK:
                self.rx_rtm_dellink(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_NEWADDR:
                self.rx_rtm_newaddr(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELADDR:
                self.rx_rtm_deladdr(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_NEWNEIGH:
                self.rx_rtm_newneigh(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELNEIGH:
                self.rx_rtm_delneigh(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_NEWROUTE:
                self.rx_rtm_newroute(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELROUTE:
                self.rx_rtm_delroute(msg)

            else:
                self.log.warning('RXed unknown netlink message type %s' % msgtype)

        if processed:
            self.netlinkq = self.netlinkq[processed:]

    def get_all_macs(self):
        family = socket.AF_BRIDGE
        debug = nlmanager.nlpacket.RTM_GETNEIGH in self.debug

        neighbor = nlmanager.nlpacket.Neighbor(nlmanager.nlpacket.RTM_GETNEIGH, debug, self.log, self.Log.getHandlerStr() != "syslog")
        neighbor.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlmanager.nlpacket.NLM_F_DUMP
        neighbor.body = struct.pack('Bxxxii', family, 0, 0)
        neighbor.build_message(self.sequence.next(), self.pid)

        if debug:
            self.debug_seq_pid[(neighbor.seq, neighbor.pid)] = True

        self.tx_nlpacket_get_response(neighbor)

    def get_all_br_links(self):
        family = socket.AF_BRIDGE
        debug = nlmanager.nlpacket.RTM_GETLINK in self.debug

        link = nlmanager.nlpacket.Link(nlmanager.nlpacket.RTM_GETLINK, debug, self.log, self.Log.getHandlerStr() != "syslog")
        link.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlmanager.nlpacket.NLM_F_DUMP
        link.body = struct.pack('Bxxxiii', family, 0, 0, 0)
        link.build_message(self.sequence.next(), self.pid)

        if debug:
            self.debug_seq_pid[(link.seq, link.pid)] = True

        self.tx_nlpacket_get_response(link)

